unit IPMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Spin, Menus;

type
  TMainForm = class(TForm)
    ExitBtn: TButton;
    ControlTogBtn: TButton;
    ForceTogBtn: TButton;
    ImpulseBtn: TButton;
    Bevel1: TBevel;
    Bevel2: TBevel;
    SpinEditImpL: TSpinEdit;
    Label5: TLabel;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Bevel3: TBevel;
    Label6: TLabel;
    Bevel4: TBevel;
    Display: TPaintBox;
    GraphDisplay: TPaintBox;
    SpinEditCartVg: TSpinEdit;
    SpinEditCartPg: TSpinEdit;
    SpinEditBobAg: TSpinEdit;
    SpinEditBobVg: TSpinEdit;
    SpinEditBobPg: TSpinEdit;
    BobGain: TLabel;
    CartGain: TLabel;
    CartAcc: TLabel;
    Bevel5: TBevel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    Bevel6: TBevel;
    ResetBtn: TButton;
    Label10: TLabel;
    MainMenu: TMainMenu;
    Help1: TMenuItem;
    AboutInvtPend1: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure DisplayMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ExitBtnMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ControlTogBtnClick(Sender: TObject);
    procedure ForceTogBtnClick(Sender: TObject);
    procedure ImpulseBtnClick(Sender: TObject);
    procedure DisplayMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure DisplayMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SpinEditImpLChange(Sender: TObject);
    procedure SpinEditCartVgChange(Sender: TObject);
    procedure SpinEditCartPgChange(Sender: TObject);
    procedure SpinEditBobAgChange(Sender: TObject);
    procedure SpinEditBobVgChange(Sender: TObject);
    procedure SpinEditBobPgChange(Sender: TObject);
    procedure ResetBtnClick(Sender: TObject);
    procedure AboutInvtPend1Click(Sender: TObject);
    procedure FormPaint(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure IdleLoop(Sender: TObject; var Done: boolean);
  end;

var
  MainForm: TMainForm;
  MouseButtonDown: Boolean;

implementation

uses IPAbout;

{$R *.dfm}
{$J+}

type ObjectType = record x, y, fx, fy, vx, vy, Mass: double end;
     CsysType = record p, r, g, e, o:double; end;

const
  XScaleFactor = 200.0;
  YScaleFactor = 200.0;

 var Bob,Cart: ObjectType;
     CartPos, CartVel, BobPos, BobVel, BobAccel: CsysType;
     DeltaX, DeltaY, dt, ForceLevel: double;
     L, L0, f, g, ke: double;
     Iters: longint;
     HCenter, VCenter, gHCenter, gVCenter: integer;
     MouseX, MouseY: integer;
     Control: boolean;
     TimeVar, ImpulseLength: double;
     ImpulseOn: boolean;

{----- World-to/from-Screen Mapping Functions ----}

function XtoScreenX(X: double): integer;
begin
  Result := HCenter + round(X * XScaleFactor);
end;

function YtoScreenY(Y: double): integer;
begin
  Result := VCenter - round(Y * YScaleFactor);
end;

function ScreenXtoX(ScreenX: integer): double;
begin
  Result := ScreenX/XScaleFactor;
end;

{----- Simulation Initialization -----}

procedure InitPendulum;
begin
  dt := 1e-4;   { TimeVar increment per iteration, sec }
  ke := 1e7;    { Rod compression force, newtons/meter of compression}
  L0 := 1.0;    { Rod-length, meters }
  g  := 9.8;    { Gravity, meter/sec^2 acceleration}
  Iters := 0;
  Control := true;
  ImpulseOn := false;
  with Bob do
    begin
      vx := 0.0;  { m/sec }
      vy := 0.0;
//      x := L0*sin(0);
//      y := L0*cos(1);  { meter }
      x := 0;
      y := L0;
      Mass := 0.5; { kg }
    end;
  with Cart do
    begin
      vx := 0.0;
      vy := 0.0;
      x := 0.0;
      y := 0.0;
      Mass := 0.5; { kg }
    end;
end;

procedure InitControlSystems;
begin
  with CartVel do
    begin
      g := 100.0;
      r :=   0.0;
      e :=   0.0;
      o :=   0.0;
    end;
  with CartPos do
    begin
      g := 50.0;
      r :=  0.0;
      e :=  0.0;
      o :=  0.0;
    end;
  with BobAccel do
    begin
      g := 50.0;
      r :=  0.0;
      e :=  0.0;
      o :=  0.0;
    end;
  with BobVel do
    begin
      g :=  0.5;
      r :=  0.0;
      e :=  0.0;
      o :=  0.0;
    end;
  with BobPos do
    begin
      g := 1.0;
      r := 0.0;
      e := 0.0;
      o := 0.0;
    end;
end;

{----- Control Systems and Dynamics -----}

procedure Dynamics;
begin
  with Bob do
    begin
      DeltaX := Bob.x - Cart.x;
      DeltaY := Bob.y - Cart.y;
      L := sqrt(sqr(DeltaX) + sqr(DeltaY)); //Computed length of pole
      f := (L0 - L)*ke; //Force proportional to pole compression
      fx := f*DeltaX/L + ForceLevel; //Horizontal force on bob
      fy := f*DeltaY/L - Mass*g; //Vertical force on bob
      vx := vx + fx/Mass* dt; //Horizontal velocity
      x := x + (vx + 0.5*fx*dt/Mass)*dt; //Horizontal position
      vy := vy + fy/Mass* dt; //Vertical velocity
      y := y + (vy + 0.5*fy*dt/Mass)*dt; //Vertical position
    end;
  with Cart do
    begin
      if Control then fx := -Bob.fx + CartVel.o //Horizontal force on cart
        else fx := -Bob.fx;
      vx := vx + fx/Mass*dt - vx*dt; //Horizontal velocity; vx*dt = damping
      x := x + (vx + 0.5*fx/Mass*dt)*dt; //Horizontal position
      if x > 3.0 then { Limit cart travel to +- 3 meters }
        begin
          fx := 0.0;
          vx := 0.0;
          x  := 3.0;
        end
        else if x < -3.0 then
          begin
            fx :=  0.0;
            vx :=  0.0;
            x  := -3.0;
         end;
    end;
end;

procedure CartVelCont;
begin
  with CartVel do
    begin
      p := Cart.vx;
      r := CartPos.o; //Proportional to cart position error
      e := r - p;
      o := o + (g*e - o)/2.0;
    end;
end;

procedure CartPosCont;
begin
  with CartPos do
    begin
      p := Cart.x;
      r := BobAccel.o; //Horizontal component of g*mass, varies with rod angle
      e := r - p;
      o := g*e;
     end
end;

procedure BobAccelCont;
begin
  with BobAccel do
    begin
      p := Bob.x - Cart.x;
      r := BobVel.o; //Proportional to error in bob horiz. velocity
      e := r - p;
      o := -g*e;
    end
end;

procedure BobVelCont;
begin
  with BobVel do
    begin
      p := Bob.vx;
      r := BobPos.o; //Proportional to error in bob position
      e := r - p;
      o := o + 10.0*(g*e - o)*dt;
    end;
end;

procedure BobPosCont;
begin
  with BobPos do
    begin
//      r := r + (ScreenXtoX(MouseX) - r)/150.0;
      r := ScreenXtoX(MouseX); //Mouse position
      p := Bob.x;
      e := r - p;
      o := g*e;
      if o > 1.4 then o := 1.4
        else if o < -1.4 then o := -1.4;
    end;
end;

{----- Graphics Display -----}

procedure ShowPendulum;
const
  OldX: integer = 0;
  OldY: integer = 0;
  OldCx: integer = 0;
  OldCy: integer = 0;

  procedure DrawCart(DrawColor: integer);
  begin
    with MainForm.Display.Canvas do
      begin
        Pen.Color := DrawColor;
        { Rod }
        MoveTo(OldCx, OldCy);
        LineTo(OldX, OldY);
        { Bob }
        Ellipse(OldX - 7, OldY - 7,
          OldX + 7, OldY + 7);
        { Truck body}
        Rectangle(OldCx - 30, VCenter,
            OldCx + 30, VCenter - 15);
        { Left wheel }
        Ellipse(OldCx - 30, OldCy - 7,
                OldCx - 16, OldCy + 7);
        { Right wheel }
        Ellipse(OldCx + 30, OldCy - 7,
                OldCx + 16, OldCy + 7);
    end;
  end;
begin
  { Erase Previous Figure }
  DrawCart(clWhite);
  { Get current bob and cart positions }
  OldX  := XtoScreenX(Bob.x); {round(200*bob.x);}
  OldY  := YtoScreenY(Bob.y); {(200*bob.y);}
  OldCx := XtoScreenX(Cart.x); {round(200*cart.x);}
  OldCy := YtoScreenY(Cart.y); {(200*cart.y);}
  { Draw bob and cart in current positions }
  DrawCart(clBlack);
  with MainForm.Display do { Draw ground }
    begin
      Canvas.MoveTo(0, VCenter + 7);
      Canvas.LineTo(Width, VCenter + 7);
    end;
end;

procedure ShowRef;
const OldRef: integer = 0;
begin
  with MainForm.Display.Canvas do
    begin
      { Erase at former position }
      Pen.Color := clWhite;
      MoveTo(OldRef,25);
      LineTo(OldRef,35);
      MoveTo(OldRef-1,25);
      LineTo(OldRef-1,35);
      { Redraw at new position }
      OldRef := XtoScreenX(bobpos.r);{HCenter + MouseX;}
      Pen.Color := clRed;
      MoveTo(OldRef,25);
      LineTo(OldRef,35);
      MoveTo(OldRef-1,25);
      LineTo(OldRef-1,35);
    end;
end;

{----- Main Initialization -----}

procedure TMainForm.FormCreate(Sender: TObject);
begin
  with Display do
    begin
      Canvas.Brush.Color := clWhite;
      Canvas.Pen.Color := clBlack;
      HCenter := Width div 2;
      VCenter := (Height div 2) + 100;
    end;
  with GraphDisplay do
    begin
      Canvas.Brush.Color := clWhite;
      Canvas.Pen.Color := clBlack;
      gHCenter := Width div 2;
      gVCenter := Height div 2;
    end;
  MouseX := 0;
  MouseY := 0;
  Iters := 0;
  InitPendulum;
  InitControlSystems;
  ImpulseLength := SpinEditImpL.Value*dt*10.0;
  ShowPendulum;
  ShowRef;
  Application.OnIdle := IdleLoop;
end;



{----- Simulation Loop -----}

procedure TMainForm.IdleLoop(Sender: TObject; var Done: boolean);
const OldRef: integer = 60;
      OldBob: integer = 60;
           X: integer = 0;
           Y: integer = 0;
     OldCart: integer = 60;
        OldX: integer = 0;

  procedure ShowImpulseMark;
  begin
    with Display.Canvas do
      begin
        Pen.Color := clRed;
        X := XtoScreenX(Bob.x);
        Y := YtoScreenY(Bob.y);
        MoveTo(X-50, Y);
        LineTo(X-10, Y);
      end
  end;

  procedure HideImpulseMark;
  begin
    with Display.Canvas do
      begin
        Pen.Color := clWhite;
        MoveTo(X-50, Y);
        LineTo(X-10, Y);
      end;
  end;

  procedure UpdateDisplays;
    begin
      ShowPendulum;
      ShowRef;
      with GraphDisplay.Canvas do
        begin
          { Bob Position Reference }
          Pen.Color := clRed;
          MoveTo(OldX, OldRef);
          OldRef := gVCenter  - round(40.0*BobPos.r);
          LineTo(OldX+2, OldRef);
          { Cart Position }
          Pen.Color := clBlue;
          MoveTo(OldX, OldCart);
          OldCart := gVCenter - round(40.0*Cart.x);
          LineTo(OldX+2, OldCart);
          { Bob Position }
          MoveTo(OldX, OldBob);
          OldBob := gVCenter  - round(40.0*Bob.x);
          Pen.Color := clGreen;
          LineTo(OldX+2, OldBob);
          Inc(OldX);
        end;
      if OldX = GraphDisplay.Width then { Reached right edge, move to left }
        begin
          GraphDisplay.Canvas.FillRect(
            Rect(0, 0, GraphDisplay.Width, GraphDisplay.Height));
          Iters := 0;
          OldX := 0;
        end;
    end;

begin
  Done := false;
  Dynamics;
  if Control then
    begin
      BobPosCont;
      BobVelCont;
      BobAccelCont;
      CartPosCont;
      CartVelCont;
    end;
  if (Iters mod 500) = 0 then UpDateDisplays { every 500th iteration };

  if ImpulseOn then
    begin
      if TimeVar = ImpulseLength then ShowImpulseMark;
      if TimeVar <= 0.0 then
        begin
          HideImpulseMark;
          ImpulseOn := false;
          ForceLevel := 0.0;
        end
      else TimeVar := TimeVar - dt;
    end;
  Inc(Iters);
end;

{----- Event Handlers -----}

procedure TMainForm.DisplayMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  if  MouseButtonDown then
    begin
      MouseX := X - HCenter;
      MouseY := Y;
    end;
end;

procedure TMainForm.ExitBtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  Application.Terminate;
end;

procedure TMainForm.ControlTogBtnClick(Sender: TObject);
begin
  if Control then
    begin
      Control := false;
      ControlTogBtn.Caption :=
      'Turn &Control ON';
    end
  else
    begin
      Control := true;
      ControlTogBtn.Caption :=
        'Turn &Control OFF';
    end;
end;

procedure TMainForm.ForceTogBtnClick(Sender: TObject);
begin
if ForceLevel = 0.0 then
  begin
    ForceTogBtn.Caption := 'Turn &Force OFF';
    ForceLevel := 1.0
  end
else
  begin
    ForceTogBtn.Caption := 'Turn &Force ON';
    ForceLevel := 0.0;
  end
end;

procedure TMainForm.ImpulseBtnClick(Sender: TObject);
begin
  if ImpulseOn then Exit;
  ImpulseOn := true;
  ForceLevel := 2.0;
  TimeVar := ImpulseLength;
  ForceTogBtn.Caption := 'Turn &Force ON';
end;

procedure TMainForm.DisplayMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  MouseButtonDown := true;
end;

procedure TMainForm.DisplayMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  MouseButtonDown := false;
end;

procedure TMainForm.SpinEditImpLChange(Sender: TObject);
begin
  ImpulseLength := SpinEditImpL.Value*dt*10.0;
end;

procedure TMainForm.SpinEditCartVgChange(Sender: TObject);
begin
  CartVel.g := SpinEditCartVg.Value;
end;

procedure TMainForm.SpinEditCartPgChange(Sender: TObject);
begin
  CartPos.g := SpinEditCartPg.Value;
end;

procedure TMainForm.SpinEditBobAgChange(Sender: TObject);
begin
  BobAccel.g := SpinEditBobAg.Value;
end;

procedure TMainForm.SpinEditBobVgChange(Sender: TObject);
begin
  BobVel.g := SpinEditBobVg.Value/10.0;
end;

procedure TMainForm.SpinEditBobPgChange(Sender: TObject);
begin
  BobPos.g := SpinEditBobPg.Value/10.0;
end;

procedure TMainForm.ResetBtnClick(Sender: TObject);
begin
  Display.Canvas.FillRect(
    Rect(0,0,Display.Width, Display.Height));

  InitPendulum;
  InitControlSystems;
  SpinEditBobAg.Value := round(BobAccel.g);
  SpinEditBobVg.Value := round(BobVel.g * 10.0);
  SpinEditBobPg.Value := round(BobPos.g * 10.0);
  SpinEditCartVg.Value := round(CartVel.g);
  SpinEditCartPg.Value := round(CartPos.g);
  SpinEditImpL.Value := 500;
  MouseX := 0;
  Control := true;
  ControlTogBtn.Caption := 'Turn &Control OFF';
  ForceLevel := 0.0;
  ForceTogBtn.Caption := 'Turn &Force ON';
end;

procedure TMainForm.AboutInvtPend1Click(Sender: TObject);
begin
  AboutBox.Left := (MainForm.Width div 2) - (AboutBox.Width div 2) - 50;
  AboutBox.Top := 125;
  AboutBox.Show;
end;

procedure TMainForm.FormPaint(Sender: TObject);
begin
  Display.Canvas.FillRect(
    Rect(0,0,Display.Width, Display.Height));
  GraphDisplay.Canvas.FillRect(
    Rect(0,0,GraphDisplay.Width,GraphDisplay.Height));
end;

end.
